import struct
import typing
import jwt
from enum import IntEnum
from urllib.parse import urlparse

# flake8: noqa: I005
import pyodbc
from azure.core.credentials import AccessToken, TokenCredential
from azure.storage.filedatalake import DataLakeDirectoryClient

class ItemType(IntEnum):
    SQL = 0
    Lakehouse_Sql = 2
    DataWarehouse = 3
    Lakehouse_Files = 4

class FabricItem:
    def __init__(self, alias_name: str, fabric_item_type: ItemType, endpoints: typing.Dict[str, typing.Dict[str, str]]):
        self.__alias_name = alias_name
        self.__fabric_item_type = fabric_item_type
        self.__endpoints = endpoints

    @property
    def alias_name(self) -> typing.Optional[str]:
        return self.__alias_name
    
    @property
    def fabric_item_type(self) -> typing.Optional[ItemType]:
        return self.__fabric_item_type
    
    @property
    def endpoints(self) -> typing.Optional[typing.Dict[str, typing.Dict[str, str]]]:
        return self.__endpoints

class FabricSqlConnection(FabricItem):

    APPSETTINGS_PATH = "sqlendpoint"

    def get_split_connection_string(self) -> typing.Dict[str, str]:

        connString = self.endpoints[self.APPSETTINGS_PATH]["ConnectionString"]

        # Lakehouse connection string contains Data Source instead of Server
        connString = connString.replace("Data Source", "Server")

        if "=" not in connString:
            return { "Server": connString }
        
        if "Server" not in connString:
            raise Exception("Server value is not set in connection")

        split_by_semicolon = connString.split(";")
        return {x.split("=")[0].strip(): x.split("=")[1].strip() for x in split_by_semicolon}

    def connect(self) -> pyodbc.Connection:
        if self.APPSETTINGS_PATH not in self.endpoints:
            raise Exception(f"{self.APPSETTINGS_PATH} is not set")
        
        dict_conn_string = self.get_split_connection_string()
        connString = dict_conn_string["Server"]

        # https://github.com/AzureAD/azure-activedirectory-library-for-python/wiki/Connect-to-Azure-SQL-Database
        
        token = self.endpoints[self.APPSETTINGS_PATH]["AccessToken"].encode('UTF-8')
        exptoken = b""
        for i in token:
            exptoken+=bytes({i})
            exptoken+=bytes(1)
        tokenstruct = struct.pack("=i", len(exptoken)) + exptoken

        driver_names = [x for x in pyodbc.drivers() if x.endswith(' for SQL Server')]
        latest_driver = driver_names[-1] if driver_names else None

        if latest_driver is None:
            raise Exception("No ODBC Driver found for SQL Server")

        return pyodbc.connect(f'DRIVER={{{latest_driver}}};Server={connString};Encrypt=yes;', attrs_before = {1256:tokenstruct}, timeout=60)
    
class FabricLakeHouseFilesClient(FabricItem):
    class CustomTokenCredential(TokenCredential):
        def __init__(self, token: str):
            alg = jwt.get_unverified_header(token)['alg']
            decodedAccessToken = jwt.decode(token, algorithms=[alg], options={"verify_signature": False, # Needed or else it complains about Exception: ValueError: Unable to load PEM file.
                                                                               "verify_exp": False} # Needed to not throw if the access token is already expired
                                                                               )
            # Token Expiry
            tokenExpiry = decodedAccessToken["exp"]

            self._token = AccessToken(token, tokenExpiry)

        def get_token(self, *scopes, **kwargs) -> AccessToken:
            return self._token
    
    APPSETTINGS_PATH = "fileendpoint"
    
    def connect(self) -> DataLakeDirectoryClient:
        if self.APPSETTINGS_PATH not in self.endpoints:
            raise Exception(f"{self.APPSETTINGS_PATH} is not set")
        
        raw_path = self.endpoints[self.APPSETTINGS_PATH]['ConnectionString']
        parsed_path = urlparse(raw_path)
        
        accessToken = self.endpoints[self.APPSETTINGS_PATH]['AccessToken']

        # The account URL is the scheme and netloc parts of the parsed path
        account_url = f"{parsed_path.scheme}://{parsed_path.netloc}"

        # The file system name and directory name are in the path part of the parsed path
        # We remove the leading slash and then split the rest into the file system name and directory name
        file_system_name, _, directory_name = parsed_path.path.lstrip('/').partition('/')

        directory_client = DataLakeDirectoryClient(account_url, file_system_name, directory_name, self.CustomTokenCredential(accessToken))
        return directory_client
    
class FabricLakehouseClient(FabricItem):

    def connectToSql(self) -> pyodbc.Connection:
        return FabricSqlConnection(self.alias_name, self.fabric_item_type, self.endpoints).connect()  

    def connectToFiles(self) -> DataLakeDirectoryClient:
        return FabricLakeHouseFilesClient(self.alias_name, self.fabric_item_type, self.endpoints).connect()

class UserDataFunctionContext:
    def __init__(self, invocationId: str, executingUser: typing.Dict[str, typing.Dict[str, str]]):
        self.invocation_id = invocationId
        self.executing_user = executingUser



# flake8: noqa: I005
from typing import Any, Callable, Dict, List, Optional, Union, \
    Iterable
from azure.functions.decorators.http import HttpTrigger, HttpOutput, \
    HttpMethod
from azure.functions.decorators.core import DataType, \
    AuthLevel, InputBinding
from azure.functions.decorators.utils import parse_singular_param_to_enum, \
    parse_iterable_param_to_enums
from azure.functions import FunctionApp, HttpResponse
from azure.functions.decorators.core import DataType
from pydoc import locate
from .middleware import RemoveReqExtension, InvocationIdMidleware
import json
import azure
import inspect

from .udf_binding import UdfPropertyInput, UdfResponse
from .fabric_class import FabricSqlConnection, FabricLakeHouseFilesClient, FabricLakehouseClient, UserDataFunctionContext
from .item_binding import FabricItemInput, ItemType
from .user_data_function_context_binding import UserDataFunctionContextInput
from functools import wraps
from typing import Callable, TypeVar
T = TypeVar('T')

class FabricApp(FunctionApp):

    def __init__(self):
        super().__init__(AuthLevel.ANONYMOUS)

    def function(self, name=None):
        @self._configure_function_builder_with_func
        def wrap(fb, user_func):
            # Add HTTP Trigger
            fb.add_trigger(trigger=HttpTrigger(
                        name='req',
                        methods=[HttpMethod.POST],
                        auth_level=AuthLevel.ANONYMOUS,
                        route=name))
            fb.add_binding(binding=HttpOutput(name='$return'))

            # Add Custom UDF Bindings
            sig = inspect.signature(user_func)
            for param in sig.parameters.values():
                if self._is_typeof_fabricitem_input(param.annotation):
                    continue
                if(param.name == 'req'):
                    continue
                if self._is_typeof_userdatafunctioncontext_input(param.annotation):
                    # fb.add_binding(binding=UserDataFunctionContext())
                    continue

                # Add custom bindings
                fb.add_binding(binding=UdfPropertyInput(
                    name=param.name, parameterName=param.name, typeName=self._get_cleaned_type(param).__name__)
                )

            return fb
        return wrap

    def _is_typeof_fabricitem_input(self, obj):
        # Check to see if parameter is anything we might return from a fabric binding
        return obj == FabricSqlConnection or obj == FabricLakeHouseFilesClient or obj == FabricLakehouseClient
    
    def _is_typeof_userdatafunctioncontext_input(self, obj):
        # Check to see if parameter is anything we might return from a fabric binding
        return obj == UserDataFunctionContext

    # If this is a data type with a nested type (e.g., list[int], dict[str, int]), we return the outermost type, e.g., list, dict
    # Otherwise, Azure Functions will say something like:
    #   Exception: FunctionLoadError: cannot load the hello_fabric2 function: binding intStrDict has invalid non-type annotation dict[int, str]
    def _get_cleaned_type(self, param):
        if(hasattr(param.annotation,'__origin__')): 
            return param.annotation.__origin__
        else:
            return param.annotation

    def _ensure_http_returntype(self, func: Callable[..., T]):
        @wraps(func)
        def wrapper(*args, **kwargs):
            resp = func(*args, **kwargs)
            if type(resp) is HttpResponse or issubclass(type(resp), HttpResponse):
                return resp
            
            return HttpResponse(str(resp))
        return wrapper

    def _configure_function_builder_with_func(self, wrap) -> Callable[..., Any]:   
        def decorator(func):
            sig = inspect.signature(func)

            # Update function parameters to include a request object for validation
            params = []
            params.append(inspect.Parameter('req', inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=azure.functions.HttpRequest))

            for param in sig.parameters.values():
                if param.name != 'req':
                    params.append(inspect.Parameter(param.name, inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=self._get_cleaned_type(param)))

            sig = sig.replace(parameters=tuple(params)).replace(return_annotation=str)
            func.__signature__ = sig

            annotations = {}
            # Update annotations to ensure it uses the cleaned type
            for param in params:
                annotations[param.name] = param.annotation

            # Update return annotation of func to be HttpResponse
            annotations['old_return'] = func.__annotations__['return']
            annotations['return'] = HttpResponse

            func.__annotations__ = annotations

            # Add wrapper for function to handle ensure all return values are parsed to HttpResponse
            user_func = self._ensure_http_returntype(func)

            fb = self._validate_type(user_func)
            self._function_builders.append(fb)

            return wrap(fb, user_func)

        return decorator
    # The decorator that will be used to tell the function we want a fabric item
    def fabric_item_input(self,
                        argName,
                        alias: str,
                        item_type: Optional[ItemType] = ItemType.SQL,
                        default_item_name: Optional[str] = None,
                        default_workspace_name: Optional[str] = None,
                        data_type : Optional[DataType] = DataType.STRING,
                        **kwargs) \
            -> Callable[..., Any]:

        @self._configure_function_builder
        def wrap(fb):
                
            fb.add_binding(
                binding=FabricItemInput(
                    name=argName,
                    alias=alias,
                    itemType=item_type,
                    defaultItemName=default_item_name,
                    defaultWorkspaceName=default_workspace_name,
                    data_type=parse_singular_param_to_enum(data_type,
                                                        DataType),
                    **kwargs))
            return fb
        
        return wrap
    
    def user_data_function_context_input(self,
                        argName,
                        **kwargs) \
            -> Callable[..., Any]:

        @self._configure_function_builder
        def wrap(fb):
            fb.add_binding(
                binding=UserDataFunctionContextInput(
                    name=argName,
                    **kwargs))
            return fb
        
        return wrap


import typing
# flake8: noqa: I005
from logging import Logger

from azure.functions import AppExtensionBase, Context


class InvocationIdMidleware(AppExtensionBase):
    """A Python worker extension to add the invocation id to the response headers.
    """

    @classmethod
    def init(cls):
        pass

    @classmethod
    def configure(cls, *args, append_to_http_response:bool=False, **kwargs):
        pass

    @classmethod
    def pre_invocation_app_level(
        cls, logger: Logger, context: Context,
        func_args: typing.Dict[str, object],
        *args, **kwargs
    ) -> None:
        pass

    @classmethod
    def post_invocation_app_level(
        cls, logger: Logger, context: Context,
        func_args: typing.Dict[str, object],
        func_ret: typing.Optional[object],
        *args, **kwargs
    ) -> None:
        if func_ret.headers is not None:
            func_ret.headers.add('x-ms-invocation-id', context.invocation_id)


# flake8: noqa: I005
import datetime
from typing import Any, Callable, Dict, List, Optional, Union, \
    Iterable
from azure.functions.decorators.core import DataType, InputBinding
import azure
from pydoc import locate
from collections import namedtuple
import json
from enum import IntEnum 
from azure.functions import HttpResponse

class UdfResponse(HttpResponse):
    pass

class UdfPropertyInput(InputBinding):
    @staticmethod
    def get_binding_name() -> str:
        return 'UdfProperty'
    
    def __init__(self,
                name: str,
                parameterName: str,
                typeName: Optional[str] = None,
                data_type: Optional[DataType] = DataType.STRING,
                **kwargs):
        super().__init__(name, data_type)


# The input converter that automatically gets registered in the function app.
class UdfPropertyConverter(azure.functions.meta.InConverter, binding='UdfProperty'):
    def try_convert_tuple(value: object) -> tuple:
        if isinstance(value, str):
            # this is the case if the tuple is passed as a string e.g., '[1,2,3]'
            value_as_json = json.loads(value)
            return tuple(value_as_json)

        return tuple(value)
    
    def try_convert_set(value: object) -> tuple:
        if isinstance(value, str):
            # this is the case if the set is passed as a string e.g., '[1,2,3]'
            value_as_json = json.loads(value)
            return set(value_as_json)

        return set(value)

    special_types = {
        "datetime": datetime.datetime.fromisoformat,
        "list": json.loads,
        "dict": json.loads,
        "set": try_convert_set,
        "tuple": try_convert_tuple
    }

    @classmethod
    def check_input_type_annotation(cls, pytype: type) -> bool:
        return True

    @classmethod
    def decode(cls, data, *,
               trigger_metadata) -> Any:
        if data is not None and data.type == 'string' and data.value is not None:
            body = json.loads(data.value)
        else:
            raise NotImplementedError(
                'unable to load data successfully for udf property')
        return tryconvert(body['PropertyType'], body['PropertyJsonValue'])


def tryconvert(property_type: str, property_json_value: str):
    if not property_json_value:  # This can happen if/when the value is empty, so we should return None
        return None
    value_as_json = json.loads(property_json_value)
    if not value_as_json:
        return None
    
    prop_type = locate(property_type)
    try:
        if prop_type.__name__ in UdfPropertyConverter.special_types:
            constructor = UdfPropertyConverter.special_types[prop_type.__name__]
            return constructor(value_as_json)
        return prop_type(value_as_json)
    except (ValueError, TypeError, json.JSONDecodeError): 
        # We can get ValueError, JSONDecodeError, and TypeError due to the special_types that have json.loads
        pass

    return value_as_json




import json
from typing import Any, Callable, Optional

# flake8: noqa: I900
import azure
from azure.functions.decorators.core import DataType, InputBinding
from azure.functions.decorators.utils import parse_singular_param_to_enum

from .fabric_class import (FabricItem, FabricLakeHouseFilesClient,
                           FabricSqlConnection, ItemType, FabricLakehouseClient)

# The binding object that will be used by our input decorator below
class FabricItemInput(InputBinding):
    @staticmethod
    def get_binding_name() -> str:
        return 'FabricItem'
    
    def __init__(self,
                name: str,
                alias: str,
                itemType: ItemType,
                defaultItemName: Optional[str] = None,
                defaultWorkspaceName: Optional[str] = None,
                data_type: Optional[DataType] = DataType.STRING,
                **kwargs):
        super().__init__(name, data_type)

# The input converter that automatically gets registered in the function app. 
class FabricItemConverter(azure.functions.meta.InConverter,
                              binding='FabricItem'):

    @classmethod
    def check_input_type_annotation(cls, pytype: type) -> bool:

        # Need to check what the enum type of the item is and make sure we can convert it
        return issubclass(pytype, (FabricItem, FabricSqlConnection))
    
    @classmethod
    def parseType(self, body: json):

        endpoints = body['Endpoints']
        endpoints = {k.lower():v for k,v in endpoints.items()} # we can know this doesn't have collisions 
        # because the host extension uses a case insensitive dictionary
        sqlEndpoint = "sqlendpoint"
        fileEndpoint = "fileendpoint"

        itemType = ItemType(body['FabricItemType'])

        if sqlEndpoint in endpoints and fileEndpoint in endpoints:
            return FabricLakehouseClient(
                alias_name=body['AliasName'],
                fabric_item_type=itemType,
                endpoints=endpoints)
        elif sqlEndpoint in endpoints:
            return FabricSqlConnection(
                alias_name=body['AliasName'],
                fabric_item_type=itemType,
                endpoints=endpoints)
        elif fileEndpoint in endpoints:
            return FabricLakeHouseFilesClient(
                alias_name=body['AliasName'],
                fabric_item_type=itemType,
                endpoints=endpoints)

        # If not found above, return the default
        return FabricItem(
                alias_name=body['AliasName'],
                fabric_item_type=itemType,
                endpoints=endpoints)
    
    @classmethod
    def decode(cls, data, *,
               trigger_metadata) -> Any:
        
        if data is not None and data.type == 'string' and data.value is not None: 
            body = json.loads(data.value) 
        else:
            raise NotImplementedError(
                f'unable to load data successfully for fabric item')
         
        body = json.loads(data.value) 

        return cls.parseType(body)


# flake8: noqa: I005
from typing import Dict, Optional
from azure.functions import AppExtensionBase, Context, HttpRequest
from logging import Logger

class RemoveReqExtension(AppExtensionBase):

    @classmethod
    def init(cls):
        cls.requests: Dict[str, HttpRequest] = {}

    @classmethod
    def pre_invocation_app_level(
        cls, logger: Logger, context: Context,
        func_args: Dict[str, object],
        *args, **kwargs
    ) -> None:
        # Remove the request object from the function arguments
        cls.requests[context.invocation_id] = func_args['req']
        del func_args['req']

    @classmethod
    def post_invocation_app_level(
        cls, logger: Logger, context: Context,
        func_args: Dict[str, object],
        func_ret: Optional[object],
        *args, **kwargs
    ) -> None:
        if context.invocation_id in cls.requests:
            func_args['req'] = cls.requests[context.invocation_id]


# flake8: noqa: I005
from typing import Any, Callable, Dict, List, Optional, Union, \
    Iterable
from azure.functions.decorators.core import  DataType
from azure.functions.decorators.core import DataType, InputBinding
import azure
from pydoc import locate
from collections import namedtuple
import json
from enum import IntEnum 
from azure.functions import HttpResponse
from .fabric_class import UserDataFunctionContext

class UserDataFunctionContextInput(InputBinding):
    @staticmethod
    def get_binding_name() -> str:
        return 'UserDataFunctionContext'
    
    def __init__(self,
                name: str,
                **kwargs):
        super().__init__(name)


# The input converter that automatically gets registered in the function app.
class UserDataFunctionContextConverter(azure.functions.meta.InConverter, binding='UserDataFunctionContext'):

    @classmethod
    def check_input_type_annotation(cls, pytype: type) -> bool:
        return pytype == UserDataFunctionContext

    @classmethod
    def decode(cls, data, *,
               trigger_metadata) -> Any:
        
        if data is not None and data.type == 'string' and data.value is not None: 
            body = json.loads(data.value) 
        else:
            raise NotImplementedError(
                f'unable to load data successfully for user Data Function Context')
         
        body = json.loads(data.value) 

        return cls.parseType(body)
    
    @classmethod
    def parseType(self, body: json):

        return UserDataFunctionContext(
                invocationId=body['InvocationId'],
                executingUser=body['ExecutingUser'])    
